// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

#include <iostream>
#include <stdlib.h>
#include <vector>
#include <list>
#include <chrono>
#include <fstream>
#include <string>
#include <filesystem>
#include <stdexcept>
using namespace std;

#include <loadbmp/loadbmp.h>
#include <Eigen/Dense>
#include <Eigen/SparseLU>
#include <Eigen/SparseCholesky>
typedef Eigen::MatrixXd Mat;
typedef Eigen::SparseMatrix<double> SpMat;
typedef Eigen::Triplet<double> Trip;
using Eigen::VectorXd;
using Eigen::MatrixXd;
using Eigen::MatrixXi;

#include "params/params.h"
#include "utils/typesdef.h"
#include "utils/math2.h"
#include "params/consts.h"
#include "params/loader.h"
#include "utils/polygon.h"

#include "utils/progress.h"

#include "solver/Tension.h"
#include "solver/Current.h"
#include "solver/Topology.h"
#include "solver/Nucleation.h"
#include "solver/Timestep.h"
#include "solver/Dynamics.h"
#include "solver/Counter.h"

#include "utils/output.h"
#include "utils/loader.h"

void capture(Configuration& configuration, Topology& topology, const char *path) {
	// Load parameters and output mask
	Parameters parameters = loadParameters(configuration);
	MatrixXi MASK = generateMask(parameters, topology);
	outputMatrixI(path, MASK, 0, 2);
}

Performances simulate(Configuration& configuration, Topology& topology, JNICaller* jniCaller) {
	// Load parameters and compute pre-valus
	Parameters parameters = loadParameters(configuration);

	// Prepare progressbar
	progressInit(jniCaller);

	// Electrical simulation
	MatrixXi MASK = generateMask(parameters, topology);
	MatrixTension TENSION = solveTension(parameters, MASK, false);
	MatrixXd CURRENT = computeCurrent(parameters, MASK, TENSION);
	MatrixXd SOURCES = computeSourceNormalized(parameters, MASK, TENSION);
	vector<Site> Sites = buildNucleationSite(parameters, topology, SOURCES);

	// Bubble dynamics initialization
	TimeSteps timesteps = {0, 0, 0, 0};
	ListBubbles Bubbles; Bubbles.flagAttached = false; Bubbles.flagDetached = false;
						 Bubbles.countAttached = 0; Bubbles.countDetached = 0;
						 Bubbles.tNextNucleation = 0;
	vector<Wall> Walls = generateWalls(parameters, topology);
	ElectrolysisStatistics Statistics = {{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, 0.0, 0.0, {0.0, 0.0}, {0.0, 0.0}};
	updateStatistics(Statistics, TENSION, CURRENT);

	// Output electrical (only if not called from jni)
	if (outputElectrical && (jniCaller == nullptr)) {
		outputMatrixI(outputFolder, "/mask.bmp", MASK, -1, 1);
		outputMatrix(outputFolder, "/tension.bmp", TENSION.VALUES, TENSION.min, TENSION.max, -1.0);
		outputData(outputFolder, "/tension.csv", parameters, TENSION.VALUES, TENSION.min, TENSION.max);
		outputMatrix(outputFolder, "/current.bmp", CURRENT, -CURRENT.maxCoeff(), CURRENT.maxCoeff(), -1.0);
		outputPoints(outputFolder, "/source.bmp", SOURCES, MAX(-SOURCES.minCoeff(), SOURCES.maxCoeff()), 7);
		outputSites(outputFolder, "/sites.bmp", parameters, Sites);
	}
	// Bubble dynamics simulation
	double t = 0.0; bool stopFlag = false;
	while ((t < parameters.tf) && (!stopFlag)) {
		// 1| process nucleation
		nucleateBubble(Bubbles, Sites, t);

		// 2| process bubble evolution
		computeBubble(parameters, Bubbles, Walls, Statistics, t, timesteps.dt, &stopFlag);

		// 3| update tension
		if (electricalCoupling) if (t > TENSION.tNext) {
			MatrixXi MASK_new = updateMask(parameters, MASK, Bubbles);
			if (MASK != MASK_new) {
				MASK = MASK_new;
				MatrixTension TENSION_new = solveTension(parameters, MASK, true);
				if (TENSION_new.sucess) {
					TENSION = TENSION_new;
					CURRENT = computeCurrent(parameters, MASK, TENSION);
					SOURCES = computeSourceNormalized(parameters, MASK, TENSION);
					updateNucleationSite(parameters, SOURCES, Sites);
					updateStatistics(Statistics, TENSION, CURRENT);
				}
				TENSION.tNext += timesteps.dtElec;
			}
		}

		// 4| update timestep
		t += timesteps.dt; timesteps.i++;
		timesteps = updateTimestep(timesteps, Bubbles.maxRadius, Bubbles.maxVelocity, Sites, &stopFlag);

		// 5| output
		progressRefresh1((t/parameters.tf), jniCaller);
		if (jniCaller == nullptr) {
			// Frame output
			outputFrame(outputFolder, "/frame.blb", t, parameters, timesteps, topology, Bubbles, Statistics);
			// Electrical outputs
			if (outputElectrical && electricalCoupling) {
				timesteps.iElec++;
				outputMatrixI(outputFolder, "/mask/", timesteps.iElec, MASK, -1, 1);
				outputMatrix(outputFolder, "/tension/", timesteps.iElec, TENSION.VALUES, TENSION.min, TENSION.max, -1.0);
				outputMatrix(outputFolder, "/current/", timesteps.iElec, CURRENT, -CURRENT.maxCoeff(), CURRENT.maxCoeff(), -1.0);
				outputPoints(outputFolder, "/source/", timesteps.iElec, SOURCES, MAX(-SOURCES.minCoeff(), SOURCES.maxCoeff()), 7);
				outputSites(outputFolder, "/sites/", timesteps.iElec, parameters, Sites);
			}
		}
	}

	// Performances
	Performances performances;
	if (!stopFlag) {
		// Compute performances
		performances = computePerformances(parameters, Statistics, Bubbles);
	} else {
		// Unable to compute performances
		performances = zeroPerformances();
	}

	return performances;
}

