// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

double str2bool(string value) {
	return value == "true";
}
int str2int(string value) {
	size_t pos;
	int intValue = stoi(value, &pos);
	if (pos != value.length()) throw invalid_argument("null");
	return intValue;
}
double str2double(string value) {
	size_t pos;
	double doubleValue = stod(value, &pos);
	if (pos != value.length()) throw invalid_argument("null");
	return doubleValue;
}

void loadParamsFromFile(Configuration& configuration) {
	ifstream file("params.txt");
    if (file.is_open()) {
        string line;
        while (getline(file, line)) {
        	if (line.empty() || line[0] == '#') continue;
        	// Find '='
        	size_t pos = line.find("=");
        	if (pos == string::npos) {
        		cerr << "Error while parsing params.txt (" << line << ")" << endl << flush;
            	exit(1);
            }
        	// Get variable and value
        	string variable = line.substr(0, pos), value = line.substr(pos + 1);

        	// Update
        	try {
            	if (variable == "dSite") dSite = str2double(value);
            	else if (variable == "RFritz") {RFritz[0] = str2double(value); RFritz[1] = str2double(value);}
            	else if (variable == "RFritz[0]") RFritz[0] = str2double(value);
            	else if (variable == "RFritz[1]") RFritz[1] = str2double(value);
            	else if (variable == "theta") {theta[0] = str2double(value); theta[1] = str2double(value);}
            	else if (variable == "theta[0]") theta[0] = str2double(value);
            	else if (variable == "theta[1]") theta[1] = str2double(value);
            	else if (variable == "sigma") sigma = str2double(value);
            	else if (variable == "electricalCoupling") electricalCoupling = str2bool(value);
            	else if (variable == "meshN") meshN = str2int(value);
            	else if (variable == "outputElectrical") outputElectrical = str2bool(value);
            	else if (variable == "outputFolder") outputFolder = value;
            	else if (variable == "useSubProgressBar") useSubProgressBar = str2bool(value);
            	else if (variable == "J") configuration.J = str2double(value);
            	else if (variable == "tf") configuration.tf = str2double(value);
            	else throw invalid_argument("null");
        	} catch (const invalid_argument& e) {
        		cerr << "Error while parsing params.txt (" << line << ")" << endl << flush;
            	exit(1);
        	}
        }
    }
}

// Dynamic pre-computation
struct Parameters {
	double L[2]; /* = dynamic precomputation */ // [m]
	int N[2]; /* = dynamic precomputation */
	double dL[2]; /* = dynamic precomputation */ // [m]
	double tf, tClogg; /* = dynamic precomputation */ // [s]
	double J; /* = dynamic precomputation */ // [A/m2]
	double I; /* = dynamic precomputation */ // [A]
};
Parameters loadParameters(Configuration& configuration) {
	// Load params file and static precomputation
	loadParamsFromFile(configuration);
	staticPreComputation();

	Parameters parameters;
	// Dimensions
	parameters.L[0] = configuration.L[0];
	parameters.L[1] = configuration.L[1];

	// Mesh
	if (parameters.L[0] > parameters.L[1]) {
		parameters.N[0] = meshN;
		parameters.N[1] = round(((double) meshN)*(parameters.L[1]/parameters.L[0]));
	} else {
		parameters.N[1] = meshN;
		parameters.N[0] = round(((double) meshN)*(parameters.L[0]/parameters.L[1]));
	}

	// Refine mesh
	if (electricalCoupling) {
		int refineFactor = 0;
		for (int c = 0; c < 2; c++) {
			int factor = (int) ceil(BUBBLE_MESH_COEFF * (parameters.L[c] / (2*Rmini[c])) / ((double) parameters.N[c]));
			if (factor > 1) {
				parameters.N[c] *= factor;
				refineFactor = mathMax(refineFactor, factor);
			}
		}
	}

	// Compute dL/dR
	parameters.dL[0] = configuration.L[0]/(parameters.N[0]-1);
	parameters.dL[1] = configuration.L[1]/(parameters.N[1]-1);

	// Current which correspond to the effective electrode surface (S = e x Ly)
	parameters.J = configuration.J;
	parameters.I = configuration.J*eThick*parameters.L[1];

	// Others
	parameters.tf = configuration.tf;
	parameters.tClogg = cloggedTimeFactor * (parameters.L[1] / uAscentMin);

	return parameters;
}
