// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

/* --------- Mask generation ---------- */
/* ------------------------------------ */
int checkTopologyPolySub(Topology& topology, double x, double y) {
    Point X = {x, y};
    for (int s = 0; s < 2; s++) for (int p = 0; p < topology.polygonCount[s]; p++) {
        int n = topology.polygonIndex[s][p+1] - topology.polygonIndex[s][p];
        Point *polys = new Point[n];
        for (int seg = 0; seg < n; seg++) {
        	int index = topology.polygonIndex[s][p] + seg;
        	polys[seg] = {topology.polygonDefinitions[index].x, topology.polygonDefinitions[index].y};
        }
        if (isInside(polys, n, X)) {
        	return (s == 0) ? 1 : -1;
        }
        delete [] polys;
    }
    return 0;
}
int checkTopologyPoly(int* N, Topology& topology, double x, double y) {
	double dY = 0.66666 / (N[1] - 1);
	double r = 0;
	if (r == 0) r = checkTopologyPolySub(topology, x, y + dY);
	if (r == 0) r = checkTopologyPolySub(topology, x, y - dY);
	return r;
}

Eigen::MatrixXi generateMask(Parameters& parameters, Topology& topology) {
	int* N = parameters.N;
	Eigen::MatrixXi MASK(N[0], N[1]);
	// Fill mask
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		double x, y; indexToPosition(N, i, j, &x, &y);
		MASK(i, j) = checkTopologyPoly(N, topology, x, y); // anode(1) electrolyte(0) cathode(-1) insulation(-2) anode_trapped(-3) cathode_trapped(-4)
	}
	return MASK;
}

void getAttachedBubbleLocation(double& x, double& y, double& r, list<Bubble>::iterator& it) {
	// Calcul du rayon actuel
	r = it->r + it->site->tmiss*it->site->growth;
	double dLMouillage = r*sin(M_PI - theta[it->site->type]);
	// Positionnement réel d'une bulle attachée
	x = it->site->x + it->site->nx*r - it->site->nx*dLMouillage;
	y = it->site->y + it->site->ny*r - it->site->ny*dLMouillage;
}

void subupdateMask(Parameters& parameters, MatrixXi& MASK, list<Bubble>::iterator& it, bool attached) {
	int* N = parameters.N;
	double* L = parameters.L;

	double itx = it->x, ity = it->y, itr = it->r;
	if (attached) getAttachedBubbleLocation(itx, ity, itr, it);

	int i, j; positionToIndex(N, &i, &j, itx/L[0], ity/L[1]);
	double x, y;


	int di = round(itr/parameters.dL[0]), dj = round(itr/parameters.dL[1]);
	for (int ii = -di; ii <= di; ii++) for (int jj = -dj; jj <= dj; jj++) {
		indexToPosition(N, i+ii, j+jj, &x, &y);
		if ((mathSQ((L[0]*x)-itx) + mathSQ((L[1]*y)-ity)) < mathSQ(itr)) {
			if (isBetween(i+ii, j+jj, N)) if (MASK(i+ii, j+jj) == 0)
				MASK(i+ii, j+jj) = -2;
		}
	}
}

MatrixXi updateMask(Parameters& parameters, MatrixXi MASK, ListBubbles& Bubbles) {
	// Replace all insulated and others specials with electrolyte
	int* N = parameters.N;
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		if ((MASK(i, j) == -2) || (MASK(i, j) == -3) || (MASK(i, j) == -4)) {
			MASK(i, j) = 0;
		}
	}
	// Add bubbles
	list<Bubble>::iterator it;
	for (it = Bubbles.attached.begin(); it != Bubbles.attached.end(); it++) {
		subupdateMask(parameters, MASK, it, true);
	}
	for (it = Bubbles.detached.begin(); it != Bubbles.detached.end(); it++) {
		subupdateMask(parameters, MASK, it, false);
	}

	// Remove insulated areas
	// Les bulles peuvent rendre certains noeuds d'électrolyte isolés (lapluspart du temps contre une électrode)
	// C'est noeuds sont alors au potentiel de l'électrode en question (J=0), mais il est nécessaire de les détecter
	MatrixXi VISITED(N[0], N[1]); VISITED.setZero();
	for (int i = 0; i < N[0]; i++) for (int j = 0; j < N[1]; j++) {
		// Vérifier qu'un noeud d'électrolyte relie bien l'anode à la cathode
		if (MASK(i,j) == 0) if (VISITED(i,j) == 0) {
			bool anode = false, cathode = false;
			list<Coord> toCheck = {{i,j}};
			VISITED(i,j) = 1;
			vector<Coord> checked;
			while (toCheck.size() > 0) {
				// Vérifie l'élement et propage
				list<Coord>::iterator coord = toCheck.begin();
				int ii = coord->i, jj = coord->j;
				// voisin x-
				if (ii > 0) {
					if (MASK(ii-1, jj) == 0) if (VISITED(ii-1, jj) == 0) {
						toCheck.push_back({ii-1, jj});
						VISITED(ii-1,jj) = 1;
					}
					if (MASK(ii-1, jj) == 1) anode = true;
					if (MASK(ii-1, jj) == -1) cathode = true;
				}
				// voisin x+
				if (ii < (N[0]-1)) {
					if (MASK(ii+1, jj) == 0) if (VISITED(ii+1, jj) == 0) {
						toCheck.push_back({ii+1, jj});
						VISITED(ii+1,jj) = 1;
					}
					if (MASK(ii+1, jj) == 1) anode = true;
					if (MASK(ii+1, jj) == -1) cathode = true;
				}
				// voisin y-
				int jjm = (jj > 0) ? jj-1 : (N[1]-1);
				if (MASK(ii, jjm) == 0) if (VISITED(ii, jjm) == 0) {
					toCheck.push_back({ii, jjm});
					VISITED(ii,jjm) = 1;
				}
				if (MASK(ii, jjm) == 1) anode = true;
				if (MASK(ii, jjm) == -1) cathode = true;
				// voisin y+
				int jjp = (jj < (N[1]-1)) ? jj+1 : 0;
				if (MASK(ii, jjp) == 0) if (VISITED(ii, jjp) == 0) {
					toCheck.push_back({ii, jjp});
					VISITED(ii,jjp) = 1;
				}
				if (MASK(ii, jjp) == 1) anode = true;
				if (MASK(ii, jjp) == -1) cathode = true;

				// valider l'élement testé
				toCheck.erase(coord);
				checked.push_back({ii, jj});
			}
			// Si le groupe ne connecte pas l'anode à la caothde...
			if (!(anode && cathode)) {
				for (vector<Coord>::iterator it = checked.begin(); it != checked.end(); it++) {
					if (anode && (!cathode)) MASK(it->i, it->j) = -3; // Il sera au même potentiel que l'anode
					else if ((!anode) && cathode) MASK(it->i, it->j) = -4; // ... que la cathode
					else MASK(it->i, it->j) = -2; // ou completement indéterminé
				}
			}
		}
	}
	return MASK;
}
/* ------------------------------------ */



/* ------ Generate wall list -------- */
/* ---------------------------------- */
vector<Wall> generateWalls(Parameters& parameters, Topology& topology) {
	list<Wall> WallList;
	for (int s = 0; s < 2; s++) for (int p = 0; p < topology.polygonCount[s]; p++) {
    	int n = topology.polygonIndex[s][p+1] - topology.polygonIndex[s][p];
    	for (int seg = 0; seg < n; seg++) {
    		int index1 = topology.polygonIndex[s][p] + seg;
    		int index2 = topology.polygonIndex[s][p] + ((seg+1)%n);
    	    Point P1 = {parameters.L[0]*topology.polygonDefinitions[index1].x, parameters.L[1]*topology.polygonDefinitions[index1].y};
    	    Point P2 = {parameters.L[0]*topology.polygonDefinitions[index2].x, parameters.L[1]*topology.polygonDefinitions[index2].y};
    	    Vect u = {P2.x-P1.x, P2.y-P1.y};
    	    double length = vectLength(u);
    	    u.x = u.x / length; u.y = u.y / length;
    	    Vect n = {u.y, -u.x};

    	    WallList.push_back({{P1.x, P1.y}, {P2.x, P2.y}, u, n, length});
    	    WallList.push_back({{P1.x, P1.y-parameters.L[1]}, {P2.x, P2.y-parameters.L[1]}, u, n, length});
    	    WallList.push_back({{P1.x, P1.y+parameters.L[1]}, {P2.x, P2.y+parameters.L[1]}, u, n, length});
    	}
    }
	vector<Wall> Walls(WallList.begin(), WallList.end());
	return Walls;
}
/* ---------------------------------- */



/* --------- Wall distance ---------- */
/* ---------------------------------- */
WallIndicator checkWall(Wall& wall, double x, double y, double maxDistance) {
	// Check if is inside
	Vect P1P = {x - wall.P1.x, y - wall.P1.y};
	double pos = vectProject(P1P, wall.u);
	if ((pos < -maxDistance) || (pos > wall.length+maxDistance))
		return {{0.0, 0.0}, {0.0, 0.0}, -1.0, nullptr};

	// Test wall side
	double dist; Vect n;
	if (pos < 0.0) {
		dist = vectLength(P1P);
		if (dist > maxDistance)
			return {{0.0, 0.0}, {0.0, 0.0}, -1.0, nullptr};
		n = {P1P.x/dist, P1P.y/dist};
		return {wall.P1, n, dist, nullptr};
	} else if (pos > wall.length) {
		Vect P2P = {x - wall.P2.x, y - wall.P2.y};
		dist = vectLength(P2P);
		if (dist > maxDistance)
			return {{0.0, 0.0}, {0.0, 0.0}, -1.0, nullptr};
		n = {P2P.x/dist, P2P.y/dist};
		return {wall.P2, n, dist, nullptr};
	} else {
		dist = vectProject(P1P, wall.n);
		if ((dist < 0.0) || (dist > maxDistance))
			return {{0.0, 0.0}, {0.0, 0.0}, -1.0, nullptr};
		n = wall.n;
		return {{wall.P1.x + pos*wall.u.x, wall.P1.y + pos*wall.u.y}, n, dist, &wall};
	}
}


void checkWalls(WallIndicator* indicators, vector<Wall>& Walls, double x, double y, double maxDistance) {
	for (int i = 0; i < wallMaxmimumIndicators; i++) {
		indicators[i].P = {-1, -1};
		indicators[i].distance = INF;
	}//return;

	for (vector<Wall>::iterator it = Walls.begin(); it != Walls.end(); it++) {
		// Test
		WallIndicator indicator = checkWall(*it, x, y, maxDistance);
		if (indicator.distance == -1.0)
			continue;

		// Une parois est détectée comme à portée...
		// Si c'est un point obtenu par linéarisation (&wall = nullptr) :
		// il ne doit ni déjà être dans la liste ni faire partie des extremums d'une parois déjà ajoutée
		if (indicator.wall == nullptr) {
			bool removeFlag = false;
			for (int i = 0; i < wallMaxmimumIndicators; i++) if (indicators[i].distance != INF) {
				if (indicators[i].wall == nullptr) {
					if (pointEquals(indicator.P, indicators[i].P)) {
						removeFlag  = true;
						break;
					}
				} else {
					if (pointEquals(indicator.P, indicators[i].wall->P1) || pointEquals(indicator.P, indicators[i].wall->P2)) {
						removeFlag  = true;
						break;
					}
				}
			}
			if (removeFlag) continue; // Rejeté !
		// Sinon, il convient de vérifier que les points extremums de la parois à ajouter ne font pas l'objet d'un point obtenu par linéarisation
		// Cependant, il peut y avoir plusieurs fois la parois (sur-contrainte logique)
		} else {
			for (int i = 0; i < wallMaxmimumIndicators; i++) if (indicators[i].distance != INF) {
				if (indicators[i].wall == nullptr) {
					if (pointEquals(indicator.wall->P1, indicators[i].P) || pointEquals(indicator.wall->P2, indicators[i].P)) {
						// Retrait et décallage
						for (int j = i; j < (wallMaxmimumIndicators-1); j++) {
							indicators[j].P = indicators[j+1].P;
							indicators[j].n = indicators[j+1].n;
							indicators[j].distance = indicators[j+1].distance;
							indicators[j].wall = indicators[j+1].wall;
						}
						indicators[wallMaxmimumIndicators-1].P = {-1, -1};
						indicators[wallMaxmimumIndicators-1].distance = INF;
						i--;
					}
				}
			}
		}

		// Save the result
		for (int i = 0; i < wallMaxmimumIndicators; i++) {
			if (indicator.distance < indicators[i].distance) {
				indicators[i].P = indicator.P;
				indicators[i].distance = indicator.distance;
				indicators[i].n = indicator.n;
				indicators[i].wall = indicator.wall;
				if (i == (wallMaxmimumIndicators-1)) maxDistance = indicator.distance;
				break;
			}
		}
	}
}
/* ---------------------------------- */
