// This file is part of BUDEES, a software that allows fast simulation of bubble dynamics in electrolyzers and their interactions with electrical parameters. 
// 
// Copyright (C) 2024 <CNRS-ENS Rennes-Université de Rennes-CY Cergy Paris Université-Université Paris Saclay-CNAM Paris-ENS Paris-Saclay>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/

/* Dynamics functions */

/*
 * Sphere-sphere long-range ------------------------------------
 */

// Mobility matrix parts for sphere-sphere longe-range interactions
void setMobilityMatrixSphere_I(double* M) {
	M[0] = 1; M[1] = 0;
	M[2] = 0; M[3] = 1;
}
void setMobilityMatrixSphere_J(double* M, double coeff, double sqRa, double sqRb, double rxab, double ryab) {
	double r_inv, r3_inv, r5_inv, part1, part2;
	r_inv = 1.0/sqrt(rxab*rxab + ryab*ryab); r3_inv = r_inv*r_inv*r_inv; r5_inv = r3_inv*r_inv*r_inv;

	double Sab2 = (sqRa + sqRb) / 3.0;
	part1 = (r_inv + Sab2*r3_inv); part2 = (r3_inv - 3.0*Sab2*r5_inv);

	M[0] = coeff*(part1 + part2*rxab*rxab); M[1] = coeff*part2*rxab*ryab;
	M[2] = coeff*part2*rxab*ryab; M[3] = coeff*(part1 + part2*ryab*ryab);
}

void addMobilityMatrixPlane(double* M, Bubble* it) {
	for (int kk = 0; kk < wallMaxmimumIndicators; kk++) {
		if (it->interactor.walls[kk].distance < (0.5*it->r*interactionLongMaximumRange)) {
			double Ra = it->r;
			double da = it->interactor.walls[kk].distance;
			double daInv = 1.0/da;

			double nx = it->interactor.walls[kk].n.x, ny = it->interactor.walls[kk].n.y;

			double part1 = -1.125*Ra*daInv;
			double part2 = -0.5625*Ra*daInv;
			double part3 = (part1 - part2);

			M[0] -= part2 + part3*nx*nx;
			M[1] -= part3*nx*ny;
			M[2] -= part3*nx*ny;
			M[3] -= part2 + part3*ny*ny;
		}
	}
}





/*
 * Sphere-sphere contact ------------------------------------
 */

// Rlub matrix parts for sphere-sphere contact interactions
void matrixRlubS(double* Rlub, bool add, double Ra, double Rb, double rxab, double ryab) {
	double Reqab = Ra*Rb/(Ra+Rb);
	double rabSq = (rxab*rxab + ryab*ryab);
	double epsad = MAX(SMALLEST, sqrt(rabSq) - Ra - Rb)/Reqab;
	double rabSqinv = 1.0/rabSq, epsadInv = 1.0/epsad;

	double part1 = 0.25*epsadInv;
	double part2 = 0.166666667*log(epsadInv);
	double part3 = (part1 - part2)*rabSqinv;

	#define Bab0 part2 + part3*rxab*rxab
	#define Bab1 part3*rxab*ryab
	#define Bab2 part3*rxab*ryab
	#define Bab3 part2 + part3*ryab*ryab
	if (add) {
		Rlub[0] += Bab0;
		Rlub[1] += Bab1;
		Rlub[2] += Bab2;
		Rlub[3] += Bab3;
	} else {
		Rlub[0] -= Bab0;
		Rlub[1] -= Bab1;
		Rlub[2] -= Bab2;
		Rlub[3] -= Bab3;
	}
}
// Sphere-sphere contact
// i
void addSphereSphereContact_I(double* Rij, double Ly, Bubble* it) {
	for (int kk = 0; kk < it->interactor.count; kk++) {
		Bubble *kt = it->interactor.bubbles[kk];
		if ((it->interactor.distancesSqNorm[kk] < interactionLubMaximumRangeSq) && (kt->r != 0)) {
			int kMode = it->interactor.imagedModes[kk];
			matrixRlubS(Rij, false, it->r, kt->r, kt->x-it->x, yRepeat(kt->y, Ly, kMode)-it->y);
		}
	}
}
// j
void addSphereSphereContact_J(double* Rij, double Ly, Bubble* it,  Bubble* jt, int j, int jMode) {
	matrixRlubS(Rij, true, it->r, jt->r, jt->x-it->x, yRepeat(jt->y, Ly, jMode)-it->y);
}




/*
 * Sphere-plan contact ------------------------------------
 */
// D matrix parts for sphere-plan contact
void matrixRlubW(double* Rlub, double Ra, double ra, double nx, double ny) {
	double epsa = MAX(SMALLEST, ra - Ra)/Ra;
	double epsaInv = 1.0/epsa;

	double part1 = -epsaInv;
	double part2 = -0.53333333333*log(epsaInv);
	double part3 = (part1 - part2);

	Rlub[0] += part2 + part3*nx*nx;
	Rlub[1] += part3*nx*ny;
	Rlub[2] += part3*nx*ny;
	Rlub[3] += part2 + part3*ny*ny;
}
// Sphere-plan contact
void addSpherePlanContact_I(double* Rij, Bubble* it) {
	for (int kk = 0; kk < wallMaxmimumIndicators; kk++) {
		if (it->interactor.walls[kk].distance < (0.5*it->r*interactionLubMaximumRange)) {
			matrixRlubW(Rij, it->r, it->interactor.walls[kk].distance, it->interactor.walls[kk].n.x, it->interactor.walls[kk].n.y);
		}
	}
}





/*
 * Repulsive terms ------------------------------------
 */
// Sphere-sphere
void addSphereSphereRepulsive(VectorXd& U0, Bubble* it, int i) {
	for (int jj = 0; jj < it->interactor.count; jj++) {
		Bubble *jt = it->interactor.bubbles[jj];
		// Check if k is in range and valid
		if ((it->interactor.distancesSqNorm[jj] < interactionRepulsiveRangeSq) && (jt->r != 0)) {
			double dxij = it->x-jt->x, dyij = it->y-jt->y;
			double rSq = (SQ(dxij) + SQ(dyij));
			double RrefSq = SQ(interactionRepulsiveRange);
			double RrirjSq = mathSQ(it->r + jt->r);

			double r = sqrt(rSq);
			dxij /= r; dyij /= r;

			double F = it->r*interactionRepulsiveVelocity*mathSQ((RrefSq - rSq) / (RrefSq - RrirjSq));

			U0[2*i+0] -= F*dxij;
			U0[2*i+1] -= F*dyij;
		}
	}
}





/*
 * Sphere-sphere force estimation ------------------------------------
 */
void computeSphereSphereForce(double* F, Bubble* it, Bubble* jt, int jMode, double Ly) {
	double Rlub[4] = {0, 0, 0, 0};
	matrixRlubS(Rlub, false, it->r, jt->r, jt->x-it->x, yRepeat(jt->y, Ly, jMode)-it->y);

	F[0] += Rlub[0]*jt->ux + Rlub[1]*jt->uy;
	F[1] += Rlub[2]*jt->ux + Rlub[3]*jt->uy;
}
